#!/usr/bin/env python

# Import the system library
import sys
# Import the python Argument parser
import argparse
# Import the maths library
import math as math

# A function to test whether a module
# is present.
def module_exists(module_name):
    # Using a try block will
    # catch the exception thrown
    # if the module is not 
    # available
    try:
        # Try to import module.
        __import__(module_name)
    # Catch the Import error.
    except ImportError:
        # Return false because
        # the module could not
        # be imported
        return False
    else:
        # The module was successfully
        # imported so return true.
        return True

# A class for the golden plover population model
class GoldenPloverPopModel (object):

    # A function to parse the input parameters file.
    def parseParameterFile(self, inputFile):
        # A string to store the input parameters to
        # be outputted in the output file.
        paramsStr = "## Input Parameters to the model.\n"
        # Open the input parameter file.
        parameterFile = open(inputFile, 'r')
        # Create a dictionary object to store the 
        # input parameters.
        params = dict()
        # Loop through each line of the input 
        # text file.
        for line in parameterFile:
            # Strip any white space either
            # side of the text.
            line = line.strip()
            # Add the line to the output 
            # parameters file.
            paramsStr += "# " + line + "\n"
            # Split the line on the '=' symbol
            paramVals = line.split("=", 1)
            # Find the known parameters and
            # convert the input string to the 
            # correct data type (i.e., float or int).
            if paramVals[0] == "numOfYears":
                params[paramVals[0]] = int(paramVals[1])
            elif paramVals[0] == "initalAdultPairPop":
                params[paramVals[0]] = int(paramVals[1])
            elif paramVals[0] == "winterSurvivalRate":
                params[paramVals[0]] = float(paramVals[1])
            elif paramVals[0] == "averageEggsPerPair":
                params[paramVals[0]] = float(paramVals[1])
            elif paramVals[0] == "averageFledgelingsPerPair":
                params[paramVals[0]] = float(paramVals[1])
            elif paramVals[0] == "predatorControl":
                if paramVals[1].lower() == "false":
                    params[paramVals[0]] = False
                elif paramVals[1].lower() == "true":
                    params[paramVals[0]] = True
                else:
                    print "predatorControl must be either True or False."
                    sys.exit()
            elif paramVals[0] == "numOfFledgelings":
                params[paramVals[0]] = int(paramVals[1])
            elif paramVals[0] == "numOfFledgelingsYearOld":
                params[paramVals[0]] = int(paramVals[1])
            elif paramVals[0] == "fledgelingsSurvivePredatorsCtrl":
                params[paramVals[0]] = float(paramVals[1])
            elif paramVals[0] == "fledgelingsSurvivePredatorsNoCtrl":
                params[paramVals[0]] = float(paramVals[1])
            else:
                # If parameter is not known then just store as
                # a string.
                params[paramVals[0]] = paramVals[1]
        # Return the parameters and parameters string
        return params, paramsStr  
        
    # A function in which the model is implemented.
    def runGPModel(self, params):
        # Set up some local variables - to be editted as
        # the model runs.
        numOfAdultsPairs = params['initalAdultPairPop']
        numOfFledgelingsYearOld = params['numOfFledgelingsYearOld']
        numOfFledgelings = params['numOfFledgelings']
        numOfEggs = 0
        
        # Create lists for the information to be outputted. 
        numOfAdultsPairsOut = list()
        numYearOldFledgelingsOut = list()
        numOfEggsOut = list()
        numOfFledgelingsOut = list()
        numOfFledgelingsB4PredOut = list()
        
        # The main model loop - looping through the years.
        for year in range(params['numOfYears']):
            # Append the output parameters at the start of the year.
            numOfAdultsPairsOut.append(numOfAdultsPairs)
            numYearOldFledgelingsOut.append(numOfFledgelingsYearOld)
            numOfFledgelingsOut.append(numOfFledgelings)
            
            # Get the number of pairs (assuming all adults 
            # are paired.
            numOfAdultsPairs += (numOfFledgelingsYearOld/2)
            # Set the number of year old fledgelings
            numOfFledgelingsYearOld = numOfFledgelings
            
            # Get the number of adults and fledgelings following winter.
            # Based on their winter survival rate.
            numOfAdultsPairs=int(numOfAdultsPairs*params['winterSurvivalRate'])
            numOfFledgelingsYearOld=int(numOfFledgelingsYearOld*params['winterSurvivalRate'])
            
            # Get the numbers of eggs to hatch
            numOfEggs = int(numOfAdultsPairs * params['averageEggsPerPair'])
            # Append to output list.
            numOfEggsOut.append(numOfEggs)
            
            # Get the number of new fledgelings.
            numOfFledgelings = int(numOfAdultsPairs * params['averageFledgelingsPerPair'])
            # Append to output.
            numOfFledgelingsB4PredOut.append(numOfFledgelings)
            
            # Apply fledgeling survival rate with an option to 
            # apply predator control (or not).
            if params['predatorControl']:
                # With predator control
                numOfFledgelings=int(numOfFledgelings*params['fledgelingsSurvivePredatorsCtrl'])
            else:
                # Without predator control
                numOfFledgelings=int(numOfFledgelings*params['fledgelingsSurvivePredatorsNoCtrl'])
        
        # Once the model has completed return the output variables for analysis.
        return numOfAdultsPairsOut, numYearOldFledgelingsOut, numOfEggsOut, numOfFledgelingsOut, numOfFledgelingsB4PredOut   
    
    # A function to write the results to a text file
    # for analysis or visualisation within another
    # package.
    def writeResultsFile(self, outputFile, paramStr, params, numOfAdultsPairsOut, numYearOldFledgelingsOut, numOfEggsOut, numOfFledgelingsOut, numOfFledgelingsB4PredOut):
        # Open the output file for writing.
        outFile = open(outputFile, 'w')
        # Write the input parameters (the string formed 
        # when we read the input parameters in. This is
        # useful as it will allow someone to understand
        # where these outputs came from.
        outFile.write(paramStr)
        # Write a header indicating the following is
        # the model outputs.
        outFile.write("\n\n## Output Results.\n")
        # Create a string for each row of the output
        # file. Each row presents a parameter.
        yearStrs = "Year"
        numOfAdultsStrs = "NumberOfAdultsPairs"
        numOfYearOldFledgesStrs = "NumberOfYearOldFledgelings"
        numOfFledgesStrs = "NumberOfFledgelings"
        numOfFledgesB4PredStrs = "NumberOfFledgelingsB4Preds"
        numOfEggsStrs = "NumberOfEggs"
        # Loop through each year, building the output strings.
        for year in range(params['numOfYears']):
            yearStrs += "," + str(year)
            numOfAdultsStrs += "," + str(numOfAdultsPairsOut[year])
            numOfYearOldFledgesStrs += "," + str(numYearOldFledgelingsOut[year])
            numOfFledgesStrs += "," + str(numOfFledgelingsOut[year])
            numOfFledgesB4PredStrs += "," + str(numOfFledgelingsB4PredOut[year])
            numOfEggsStrs += "," + str(numOfEggsOut[year])
        
        # Add a new line character to the end of each row.
        yearStrs += "\n"
        numOfAdultsStrs += "\n"
        numOfYearOldFledgesStrs += "\n"
        numOfFledgesStrs += "\n"
        numOfFledgesB4PredStrs += "\n"
        numOfEggsStrs += "\n"
        
        # Write the rows to the output file.
        outFile.write(yearStrs)
        outFile.write(numOfAdultsStrs)
        outFile.write(numOfFledgesStrs)
        outFile.write(numOfFledgesB4PredStrs)
        outFile.write(numOfFledgesStrs)
        outFile.write(numOfEggsStrs)
        
        # Close the output file.
        outFile.close()
    
    def plots(self, outputFile, params, numOfAdultsPairsOut, numYearOldFledgelingsOut, numOfEggsOut, numOfFledgelingsOut, numOfFledgelingsB4PredOut):        
        # Test that the matplotlib library is present 
        # so plots can be created.
        if module_exists("matplotlib.pyplot"):
            # The matplotlib library exists so
            # import it for use in this function.
            # Importing a library within a function
            # like this means that it is only
            # available within this function.
            import matplotlib.pyplot as plt
            # Get the number of years as a list
            # of years.
            years = range(params['numOfYears'])
            
            # Create a simple plot for the number of
            # pairs.
            fig1 = plt.figure(figsize=(15, 5), dpi=150)
            plt.plot(years, numOfAdultsPairsOut)
            plt.title("Number of pairs per year predicted by model")
            plt.xlabel("Year")
            plt.ylabel("Number Of Pairs")
            plt.savefig((outputFile+"_adultpairs.pdf"), format='PDF')
            
            # Create a simple plot for the number of
            # year old fledgelings 
            fig2 = plt.figure(figsize=(15, 5), dpi=150)
            plt.plot(years, numYearOldFledgelingsOut)
            plt.title("Number of year olf fledgelings predicted by model")
            plt.xlabel("Year")
            plt.ylabel("Number Of Fledglings")
            plt.savefig((outputFile+"_numYearOldFledgelings.pdf"), format='PDF')
            
            # Create a simple plot for the number of
            # eggs hatched each year.
            fig3 = plt.figure(figsize=(15, 5), dpi=150)
            plt.plot(years, numOfEggsOut)
            plt.title("Number of eggs per year predicted by model")
            plt.xlabel("Year")
            plt.ylabel("Number Of Eggs")
            plt.savefig((outputFile+"_numOfEggs.pdf"), format='PDF')
            
            # Create a simple plot for the number of
            # new born fledgelings
            fig4 = plt.figure(figsize=(15, 5), dpi=150)
            plt.plot(years, numOfFledgelingsOut)
            plt.title("Number of fledgelings per year predicted by model")
            plt.xlabel("Year")
            plt.ylabel("Number Of Fledgelings")
            plt.savefig((outputFile+"_numOfFledgelings.pdf"), format='PDF')
            
            # Create a simple plot for the number of
            # fledgelings before that years breeding
            fig5 = plt.figure(figsize=(15, 5), dpi=150)
            plt.plot(years, numOfFledgelingsB4PredOut)
            plt.title("Number of fledgelings before breeding per year predicted by model")
            plt.xlabel("Year")
            plt.ylabel("Number Of Fledgelings")
            plt.savefig((outputFile+"_numOfFledgelingsB4Pred.pdf"), format='PDF')
            
        else:
            # If the matplotlib library is not available
            # print out a suitable error message.
            print "Matplotlib is not available and therefore the plots cannot be created." 
    
    # The run function controlling the overall order
    # of when things run.
    def run(self, inputFile, outputFile, plotsPath):
        # Provide user feedback to the user.
        print "Parse Input File."
        # Call the function to parse the input file.
        params, paramsStr = self.parseParameterFile(inputFile)
        # Print he parameters.
        print params
        # Provide some progress feedback to the user
        print "Run the model"
        # Run the model and get the output parameters.
        numOfAdultsPairsOut, numYearOldFledgelingsOut, numOfEggsOut, numOfFledgelingsOut, numOfFledgelingsB4PredOut = self.runGPModel(params)
        # Provide some feedback to the user.
        print "Write the results to an output file"
        # Call the function to write the outputs
        # to a text file.
        self.writeResultsFile(outputFile, paramsStr, params, numOfAdultsPairsOut, numYearOldFledgelingsOut, numOfEggsOut, numOfFledgelingsOut, numOfFledgelingsB4PredOut)
        # Check whether a path has been provided
        # for the plots. If it has then generate
        # output plots.
        if plotsPath is not None:
            # Give the user feedback of what's happenign.
            print "Generating plots of the results"
            # Call the function to generate plots
            self.plots(plotsPath, params, numOfAdultsPairsOut, numYearOldFledgelingsOut, numOfEggsOut, numOfFledgelingsOut, numOfFledgelingsB4PredOut)

# This is the first part of the script to 
# be executed.
if __name__ == '__main__':
    # Create the command line options 
    # parser.
    parser = argparse.ArgumentParser()
    # Define the argument for specifying the input file.
    parser.add_argument("-i", "--input", type=str, help="Specify the input image file.")
    # Define the argument for specifying the output file.
    parser.add_argument("-o", "--output", type=str, help="Specify the output text file.")
    # Define the argument for specifying the output file.
    parser.add_argument("-p", "--plot", type=str, help="Specify the output base path for the plots.")
    # Call the parser to parse the arguments.
    args = parser.parse_args()
    
    # Check that the input parameter has been specified.
    if args.input == None:
        # Print an error message if not and exit.
        print "Error: No input image file provided."
        sys.exit()
        
    # Check that the input parameter has been specified.
    if args.output == None:
        # Print an error message if not and exit.
        print "Error: No output text file provided."
        sys.exit()
    
    # Create instance of the model class
    obj = GoldenPloverPopModel()
    # Call the run function to execute the model.
    obj.run(args.input, args.output, args.plot)
